# -*- coding: utf-8 -*-

from requests.exceptions import ConnectionError
import logging
import datetime

#from ...conf import RIPPLE_ACCOUNT, RIPPLE_SECRET
from django.conf import settings
from django.core.management.base import NoArgsCommand

from ripple_api.ripple_api import account_tx, tx, RippleApiError
from ripple_api.models import Transaction
from ripple_api.tasks import sign_task, submit_task


MAX_RESULTS = 200


def get_min_ledger_index():
    transactions = Transaction.objects.filter(status__in=[
            Transaction.RECEIVED, Transaction.PROCESSED,
            Transaction.MUST_BE_RETURN, Transaction.RETURNING,
            Transaction.RETURNED
            ]).order_by('-pk')
    return transactions[0].ledger_index if transactions else -1


class Command(NoArgsCommand):
    help = 'Command that processes transactions.'
    logger = logging.getLogger('ripple')
    logger.setLevel(40)

    def format_log_message(self, message, transaction=None, *args):
        if transaction or args:
            format_args = [transaction]
            format_args.extend(args)
            return message % tuple(format_args)
        else:
            return message

    def handle_noargs(self, **options):
        self.retry_failed_transactions()
        self.monitor_transactions()
        self.return_funds()
        self.submit_pending_transactions()
        self.check_submitted_transactions()

    def monitor_transactions(self):
        """
        Get new transactions for settings.RIPPLE_ACCOUNT and store them in DB

        """
        start_time = datetime.datetime.now()
        self.logger.info(self.format_log_message(
                'Looking for new ripple transactions since last run'))
        ledger_min_index = get_min_ledger_index()
        marker = None
        has_results = True
        try: timeout = settings.RIPPLE_TIMEOUT
        except AttributeError: timeout = 5

        while has_results:
            try:
                response = account_tx(settings.RIPPLE_ACCOUNT,
                                      ledger_min_index, limit=200,
                                      marker=marker,
                                      timeout=timeout)
                # self.logger.info(self.format_log_message(response))
            except (RippleApiError, ConnectionError), e:
                self.logger.error(self.format_log_message(e))
                break

            transactions = response['transactions']
            marker = response.get('marker')
            has_results = bool(marker)

            for transaction in transactions:
                tr_tx = transaction['tx']
                meta = transaction.get('meta', {})
                if meta.get('TransactionResult') != 'tesSUCCESS':
                    continue
                amount = meta.get('DeliveredAmount') or tr_tx.get('Amount', {})

                unprocessed_unstored_transactions = (
                    tr_tx['TransactionType'] == 'Payment' and
                    tr_tx['Destination'] == settings.RIPPLE_ACCOUNT and
                    isinstance(amount, dict) and
                    not Transaction.objects.filter(hash=tr_tx['hash'])
                )
                if unprocessed_unstored_transactions:
                        self.logger.info(self.format_log_message('Saving transaction: %s', transaction))
                        destination_tag = tr_tx.get('DestinationTag')
                        source_tag = tr_tx.get('SourceTag')

                        transaction_object = Transaction.objects.create(
                            account=tr_tx['Account'], hash=tr_tx['hash'],
                            destination=settings.RIPPLE_ACCOUNT,
                            ledger_index=tr_tx['ledger_index'],
                            destination_tag=destination_tag,
                            source_tag=source_tag, status=Transaction.RECEIVED,
                            currency=amount['currency'],
                            issuer=amount['issuer'], value=amount['value']
                        )
                        self.logger.info(self.format_log_message(
                                "Transaction saved: %s", transaction_object))
            if (datetime.datetime.now() - start_time
                >= datetime.timedelta(seconds=270) and has_results):
                has_results = False
                self.logger.error(
                    'Process_transactions command terminated because '
                    '(270 seconds) timeout: ' + unicode(marker))

    def check_submitted_transactions(self):
        """
        Check final disposition of transactions.
        """
        self.logger.info(self.format_log_message('Checking submitted transactions'))
        for transaction in Transaction.objects.filter(status=Transaction.SUBMITTED):
            try:
                response = tx(transaction.hash)
                self.logger.info(self.format_log_message(response))
            except RippleApiError as e:
                self.logger.error(self.format_log_message("Error processing %s: %s", transaction, e))
                if e.error == 'txnNotFound':
                    self.logger.info(self.format_log_message(
                        'Setting transaction status to Failed for %s', transaction
                    ))
                    transaction.status = Transaction.FAILURE
                    transaction.save()
                continue

            status = response.get('meta', {}).get('TransactionResult')

            if status == 'tesSUCCESS':
                transaction.status = Transaction.SUCCESS
                transaction.save()

                if transaction.parent:
                    transaction.parent.status = Transaction.RETURNED
                    transaction.parent.save()

                self.logger.info(self.format_log_message(
                    "Transaction: %s to %s was complete.", transaction, transaction.destination
                    )
                )
            else:
                self.logger.info("Transaction status: %s" % status)

    def submit_pending_transactions(self):
        """
        Submit transactions that was signed, but connection error occurred when submit it.
        """
        self.logger.info(self.format_log_message('Submit pending transactions'))
        for transaction in Transaction.objects.filter(status=Transaction.PENDING):
            self.logger.info(self.format_log_message('Submit: %s', transaction))
            submit_task.apply((transaction,))

    def return_funds(self):
        self.logger.info('Returning failed stakes')
        for transaction in Transaction.objects.filter(
                status=Transaction.MUST_BE_RETURN):
            self.logger.info("Transaction %s must be return." % transaction.pk)

            ret_transaction = Transaction.objects.create(
                account=settings.RIPPLE_ACCOUNT,
                destination=transaction.account,
                currency=transaction.currency,
                value=transaction.value,
                status=Transaction.PENDING,
                parent=transaction
            )
            sign_task.apply((ret_transaction, settings.RIPPLE_SECRET))
            transaction.status = Transaction.RETURNING
            transaction.save()
            self.logger.info("New transaction created for returning %s", ret_transaction.pk)

    def retry_failed_transactions(self):
        self.logger.info('Retrying failed transactions')
        for transaction in Transaction.objects.filter(
                status=Transaction.FAILURE):
            self.logger.info(self.format_log_message('Found %s', transaction))

            retry_transaction = Transaction.objects.create(
                account=transaction.account,
                destination=transaction.destination,
                currency=transaction.currency,
                value=transaction.value,
                status=Transaction.PENDING,
                parent=transaction.parent
            )
            sign_task.apply((retry_transaction, settings.RIPPLE_SECRET))
            self.logger.info("New transaction created for returning %s", retry_transaction.pk)
            transaction.status = Transaction.FAIL_FIXED
            transaction.save()
            self.logger.info("Fixed the transaction")



# -*- coding: utf-8 -*-
from django.test import TestCase
from django.conf import settings
from django.db.models.signals import post_save

from mock import patch
from requests import ConnectionError, Response
import json
import ssl

from .models import Transaction
from .management.commands.process_transactions import Command
from .signals import transaction_status_changed
from ripple_api import call_api, RippleApiError


class TestRipple(TestCase):

    @patch('ripple_api.ripple_api.call_api')
    def test_monitor_transactions(self, call_api_mock):
        data = {'transactions': [
            {'tx': {'Account': 'account', 'Destination': settings.RIPPLE_ACCOUNT,
                'TransactionType': 'Payment', 'Amount': {'currency': 'CCK', 'issuer': 'p2pay', 'value': '1.2'},
                'hash': 'hash1', 'ledger_index': 1, 'DestinationTag': 232 },
             'meta': {'TransactionResult': 'tesSUCCESS'}},
        ]}
        call_api_mock.return_value = data

        receivers = post_save.receivers
        post_save.receivers = []
        Command().monitor_transactions()
        post_save.receivers = receivers

        self.assertEqual(Transaction.objects.count(), 1)
        transaction = Transaction.objects.get(id=1)
        self.assertEqual(transaction.status, Transaction.RECEIVED)
        self.assertEqual(transaction.account, 'account')
        self.assertEqual(transaction.destination, settings.RIPPLE_ACCOUNT)
        self.assertEqual(transaction.currency, 'CCK')
        self.assertEqual(transaction.issuer, 'p2pay')
        self.assertEqual(transaction.value, '1.2')
        self.assertEqual(transaction.hash, 'hash1')
        self.assertEqual(transaction.destination_tag, 232)


    @patch('ripple_api.tasks.sign')
    @patch('ripple_api.tasks.path_find')
    def test_retry(self, path_find_mock, sign_mock):
        transaction = Transaction.objects.create(
            account='account',
            destination='destination',
            hash='hash',
            status=Transaction.FAILURE,
            currency='XXX',
            tx_blob='tx_blob')

        path_find_mock.return_value = {
            'alternatives': [{
                'paths_computed': [
                    [{u'account': u'rp2PaYDxVwDvaZVLEQv7bHhoFQEyX1mEx7',
                      u'type': 1,
                      u'type_hex': u'0000000000000001'}],
                ]},
            ],
        }
        sign_mock.return_value = {
            'tx_blob': 'tx_new_blob',
            'tx_json': {
                'hash': 'new_hash'
            }
        }
        Command().retry_failed_transactions()

        transaction = Transaction.objects.get(id=transaction.id)
        self.assertEqual(transaction.status, Transaction.FAIL_FIXED)

        retry_transaction = Transaction.objects.get(hash='new_hash')
        self.assertEqual(retry_transaction.hash, 'new_hash')
        self.assertEqual(retry_transaction.tx_blob, 'tx_new_blob')
        self.assertEqual(retry_transaction.status, Transaction.PENDING)
        self.assertEqual(transaction.value, retry_transaction.value)
        self.assertEqual(transaction.account, retry_transaction.account)
        self.assertEqual(transaction.currency, retry_transaction.currency)
        self.assertEqual(
            transaction.destination,
            retry_transaction.destination
        )
        self.assertEqual(transaction.source_tag, retry_transaction.source_tag)
        self.assertEqual(
            transaction.destination_tag,
            retry_transaction.destination_tag
        )

    @patch('ripple_api.tasks.sign')
    @patch('ripple_api.tasks.path_find')
    def test_return_funds(self, path_find_mock, sign_mock):
        transaction = Transaction.objects.create(
            account='account',
            hash='hash4',
            status=Transaction.MUST_BE_RETURN,
            currency='CCK',
            value='100.1'
        )
        path_find_mock.return_value = {
            'alternatives': [{
                'paths_computed': [
                    [{u'account': u'rp2PaYDxVwDvaZVLEQv7bHhoFQEyX1mEx7',
                      u'type': 1,
                      u'type_hex': u'0000000000000001'}],
                ]},
            ],
        }
        sign_mock.return_value = {
            'tx_blob': 'tx_blob',
            'tx_json': {
                'hash': 'hash5'
            }
        }

        Command().return_funds()
        returning_transaction = Transaction.objects.get(hash='hash5')

        self.assertEqual(
            returning_transaction.destination,
            transaction.account
        )
        self.assertEqual(returning_transaction.status, Transaction.PENDING)
        self.assertEqual(returning_transaction.tx_blob, 'tx_blob')

        self.assertEqual(returning_transaction.parent, transaction)

    @patch('ripple_api.ripple_api.call_api')
    def test_check_submitted_transactions(self, call_api_mock):
        transaction = Transaction.objects.create(account='account', hash='hash', status=Transaction.RETURNING)
        Transaction.objects.create(account='account', hash='hash1', status=Transaction.SUBMITTED, parent=transaction)

        data = {'meta': {'TransactionResult': 'tesSUCCESS'}}

        call_api_mock.return_value = data
        receivers = transaction_status_changed.receivers
        transaction_status_changed.receivers = []
        Command().check_submitted_transactions()
        transaction_status_changed.receivers = receivers

        transaction = Transaction.objects.get(hash='hash1')
        self.assertEqual(transaction.status, Transaction.SUCCESS)
        self.assertEqual(transaction.parent.status, Transaction.RETURNED)

    @patch('ripple_api.ripple_api.call_api')
    def test_submit_pending(self, call_api_mock):
        Transaction.objects.create(
                account='account',
                hash='hash',
                status=Transaction.PENDING,
                tx_blob='tx_blob')
        data = {'engine_result': 'tesSUCCESS'}

        call_api_mock.return_value = data
        Command().submit_pending_transactions()
        transaction = Transaction.objects.get(hash='hash')
        self.assertEqual(transaction.status, Transaction.SUBMITTED)

    @patch('requests.post')
    def test_call_api(self, post_mock):
        original_settings = settings.RIPPLE_API_DATA
        settings.RIPPLE_API_DATA = [
            {
                'RIPPLE_API_URL': 'http://one.ripple.com:51234',
                'RIPPLE_API_USER': '',
                'RIPPLE_API_PASSWORD': '',
            },
            {
                'RIPPLE_API_URL': 'http://two.ripple.com:51234',
                'RIPPLE_API_USER': '',
                'RIPPLE_API_PASSWORD': '',
            },
            {
                'RIPPLE_API_URL': 'http://three.ripple.com:51234',
                'RIPPLE_API_USER': '',
                'RIPPLE_API_PASSWORD': '',
            }
        ]

        def custom_call_api(error):
            try:
                call_api({})
            except error:
                self.assertEqual(post_mock.call_count,
                                 len(settings.RIPPLE_API_DATA))

        def side_effect(*args, **kwargs):
            raise ConnectionError

        post_mock.side_effect = side_effect

        custom_call_api(ConnectionError)

        def side_effect(*args, **kwargs):
            response = Response()
            response._content = '''{'\2': 'binary', 'result': 'failed'}'''
            return response

        post_mock.reset_mock()
        post_mock.side_effect = side_effect

        custom_call_api(RippleApiError)

        def side_effect(*args, **kwargs):
            response = Response()
            response_data = {'result': {}}
            response_data['result']['error'] = 'failed'
            response_data['result']['error_code'] = 403
            response_data['result']['error_message'] = 'You failed!'
            response._content = json.dumps(response_data)
            return response

        post_mock.reset_mock()
        post_mock.side_effect = side_effect

        custom_call_api(RippleApiError)

        settings.RIPPLE_API_DATA = original_settings

    @patch('requests.post')
    def test_timeout(self, post_mock):
        post_mock.side_effect = ssl.SSLError

        with self.assertRaises(RippleApiError):
            call_api({})



# -*- coding: utf-8 -*-
from south.utils import datetime_utils as datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'Transaction'
        db.create_table('ripple_api_transaction', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('account', self.gf('django.db.models.fields.CharField')(max_length=100)),
            ('destination', self.gf('django.db.models.fields.CharField')(max_length=100)),
            ('hash', self.gf('django.db.models.fields.CharField')(unique=True, max_length=100)),
            ('tx_blob', self.gf('django.db.models.fields.TextField')(blank=True)),
            ('currency', self.gf('django.db.models.fields.CharField')(max_length=3)),
            ('issuer', self.gf('django.db.models.fields.CharField')(max_length=100)),
            ('value', self.gf('django.db.models.fields.CharField')(max_length=100)),
            ('source_tag', self.gf('django.db.models.fields.IntegerField')(null=True, blank=True)),
            ('destination_tag', self.gf('django.db.models.fields.IntegerField')(null=True, blank=True)),
            ('ledger_index', self.gf('django.db.models.fields.IntegerField')(null=True, blank=True)),
            ('status', self.gf('django.db.models.fields.SmallIntegerField')(default=0)),
            ('parent', self.gf('django.db.models.fields.related.ForeignKey')(blank=True, related_name='returning_transaction', null=True, to=orm['ripple_api.Transaction'])),
            ('created', self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True, blank=True)),
        ))
        db.send_create_signal('ripple_api', ['Transaction'])


    def backwards(self, orm):
        # Deleting model 'Transaction'
        db.delete_table('ripple_api_transaction')


    models = {
        'ripple_api.transaction': {
            'Meta': {'object_name': 'Transaction'},
            'account': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'currency': ('django.db.models.fields.CharField', [], {'max_length': '3'}),
            'destination': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'destination_tag': ('django.db.models.fields.IntegerField', [], {'null': 'True', 'blank': 'True'}),
            'hash': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'issuer': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'ledger_index': ('django.db.models.fields.IntegerField', [], {'null': 'True', 'blank': 'True'}),
            'parent': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'returning_transaction'", 'null': 'True', 'to': "orm['ripple_api.Transaction']"}),
            'source_tag': ('django.db.models.fields.IntegerField', [], {'null': 'True', 'blank': 'True'}),
            'status': ('django.db.models.fields.SmallIntegerField', [], {'default': '0'}),
            'tx_blob': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            'value': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        }
    }

    complete_apps = ['ripple_api']


# -*- coding: utf-8 -*-
from requests import ConnectionError
from celery import task
import logging
import time

from .models import Transaction
from ripple_api import RippleApiError, path_find, sign, submit, tx


@task
def sign_task(transaction, secret):
    logger = logging.getLogger('ripple')

    try:
        if transaction.currency == 'XRP':
            amount = transaction.value
        else:
            amount = {"currency": transaction.currency, 
                      "value": str(transaction.value), 
                      "issuer": transaction.destination}
            paths = path_find(transaction.account,
                              transaction.destination,
                              amount,
                              [{'currency': transaction.currency,
                                'issuer': transaction.account,
                                },
                               ])
            if len(paths['alternatives'])==0:
                raise RippleApiError('no path', '', 
                                     'No path between %s and %s for %s' % (
                        transaction.account, transaction.destination, amount))
            pass

        response = sign(transaction.account, 
                        secret, 
                        transaction.destination, 
                        amount,
                        paths = paths['alternatives'][0]['paths_computed']
                        )
    except (RippleApiError, ConnectionError), e:
        transaction.status = Transaction.FAILURE
        transaction.save()
        logger.error(e)
        return

    transaction.hash = response['tx_json']['hash']
    transaction.tx_blob = response['tx_blob']
    transaction.status = Transaction.PENDING
    transaction.save()

    logger.info('Transaction signed: %s' % transaction)
    return transaction


@task
def submit_task(transaction):
    if transaction:
        logger = logging.getLogger('ripple')
        try:
            response = submit(transaction.tx_blob)
            if response['engine_result'] == "telINSUF_FEE_P":
                # ripple server too busy to forward or process your transaction
                while True:
                    time.sleep(2)
                    tx_response = tx(transaction.hash)
                    status = tx_response.get('meta', {}).get('TransactionResult')
                    # check if tx status is available
                    if status:
                        response['engine_result'] = status
                        break

        except RippleApiError, e:
            logger.error(e)
            transaction.status = Transaction.FAILURE
            transaction.save()
            return
        except ConnectionError, e:
            logger.error('Connection error: %s' % e)
            return

        if response['engine_result'] in ["tesSUCCESS",  "tefPAST_SEQ"]:
            transaction.status = Transaction.SUBMITTED
            transaction.save()
            logger.info("Transaction: %s successful submitted." % transaction)
        else:
            transaction.status = Transaction.FAILURE
            transaction.save()
            logger.info("Transaction: %s submitted with result %s" % (transaction, response['engine_result']))
